# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: stats
   :platform: Unix
   :synopsis: grabs a selection of statistics.

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>
"""
import logging
import numpy as np

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin

from savu.plugins.utils import register_plugin


@register_plugin
class Stats(Plugin, CpuPlugin):
    """
    :param out_datasets: the output dataset. Default: ['stats'].
    :param required_stats: create a list of required stats\
        calcs. Default: ['max'].
    :param direction: which direction to perform this. Default: 'PROJECTION'.
    """

    def __init__(self):
        logging.debug("Starting the statistics")
        super(Stats, self).__init__("Stats")

    def process_frames(self, data):
        data = data[0]
        if 'max' in self.parameters['required_stats']:
            maximum = np.max(data)
            op = np.array([maximum])
        return op

    def post_process(self):
        out_datasets = self.get_out_datasets()
        in_meta_data = self.get_in_meta_data()[0]
        data = out_datasets[0].data
        if self.parameters['required_stats'] == 'max':
            maximum = np.max(data)
            in_meta_data.set('max', maximum)

    def get_max_frames(self):
        return 'single'

    def setup(self):
        self.exp.log(self.name + " Start")
        _in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.parameters["direction"],
                                      self.get_max_frames())
        nFrames = in_pData[0].get_total_frames()
        axis_labels = ['frame.unit', 'max.unit']
        out_dataset[0].create_dataset(axis_labels=axis_labels,
                                      shape=(nFrames, 1),
                                      remove=True)

        out_dataset[0].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))
        out_pData[0].plugin_data_setup("METADATA", self.get_max_frames())



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: strip_background
   :platform: Unix
   :synopsis: A plugin to automatically strip peaks and subtract background

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""
import logging
from scipy.signal import savgol_filter
import numpy as np
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
import time
from savu.plugins.utils import register_plugin
from copy import deepcopy


@register_plugin
class StripBackground(BaseFilter, CpuPlugin):
    """
    1D background removal. PyMca magic function

    :param iterations: Number of iterations. Default: 100.
    :param window: Half width of the rolling window. Default: 10.
    :param SG_filter_iterations: How many iterations until smoothing \
        occurs. Default: 5.
    :param SG_width: Whats the savitzgy golay window. Default: 35.
    :param SG_polyorder: Whats the savitzgy-golay poly order. Default: 5.
    :param out_datasets: A list of the dataset(s) to process. Default: ['in_datasets[0]','background'].
    """

    def __init__(self):
        logging.debug("Stripping background")
        super(StripBackground, self).__init__("StripBackground")

    def process_frames(self, data):
        data = data[0]
        t1 = time.time()
        its = self.parameters['iterations']
        w = self.parameters['window']
        smoothed = self.parameters['SG_filter_iterations']
        SGwidth = self.parameters['SG_width']
        SGpoly = self.parameters['SG_polyorder']

        npts = len(data)
        x = np.arange(npts)  # set up some x indices
        #print data.shape
        filtered = savgol_filter(data, 35, 5) # make the start a bit a bit smoother
        # lets do it the crap, slow way first
        aved = np.zeros_like(filtered)
        bottomedgemain = x < w
        bottomedgerest = (x >= w) & (x < 2*w)

        mainpart = (x >= w) & (x < (npts-w))
        mainpartbottom = (x >= 0) & (x < (npts-2*w))
        mainparttop = (x >= 2*w) & (x < (npts))

        topedgemain = x >= (npts-w)
        topedgerest = (x >= (npts-2*w)) & (x >= (npts-w))

        for k in range(its):
            aved[mainpart] = (filtered[mainpartbottom] + filtered[mainpart] + filtered[mainparttop]) / 3.0
            aved[bottomedgemain] = (filtered[bottomedgemain] + filtered[bottomedgerest]) / 2.0
            aved[topedgemain] = (filtered[topedgemain] + filtered[topedgerest]) / 2.0
            filtered[aved<filtered] = aved[aved<filtered]
            if not (k/float(smoothed)-k/int(smoothed)):
                filtered=savgol_filter(filtered,35,5)

        t2 = time.time()
        logging.debug("Strip iteration took: %s ms", str((t2-t1)*1e3))
#         print (data - filtered).shape
        return [data - filtered, filtered]

    def setup(self):
        logging.debug('setting up the background subtraction')
        in_dataset, out_datasets = self.get_datasets()
        #print in_dataset, out_datasets

        in_meta = in_dataset[0].meta_data
        in_dictionary = in_meta.get_dictionary()
        #print in_dictionary
        stripped = out_datasets[0]
        stripped.create_dataset(in_dataset[0])
        stripped.meta_data.dict = deepcopy(in_dictionary)
        #print stripped.meta_data.dict
        background = out_datasets[1]
        background.create_dataset(in_dataset[0])
        background.meta_data.dict = deepcopy(in_dictionary)

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[1].plugin_data_setup('SPECTRUM', self.get_max_frames())

        logging.debug("****STRIP_BACKGROUND AXIS LABELS*** %s",
                      stripped.get_axis_labels())

    def get_max_frames(self):
        return 'single'

    def nOutput_datasets(self):
        return 2



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: denoise_bregman_filter
   :platform: Unix
   :synopsis: A plugin to denoise 2D slices of data by using the Split Bregman
              to solve the Total Variation ROF model

.. moduleauthor:: Imanol Luengo <scientificsoftware@diamond.ac.uk>
"""
import logging
import numpy as np

from skimage.restoration import denoise_tv_bregman

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin

from savu.plugins.utils import register_plugin


@register_plugin
class DenoiseBregmanFilter(BaseFilter, CpuPlugin):
    """
    Split Bregman method for solving the denoising Total Variation ROF model.

    :param weight: Denoising factor. Default: 2.0
    :param max_iterations: Total number of iterations. Default: 100.
    :param error_threshold: Convergence threshold. Default: 0.001.
    :param isotropic: Isotropic or Anisotropic filtering. Default: False.
    """

    def __init__(self):
        logging.debug("Starting Denoise Bregman Filter")
        super(DenoiseBregmanFilter, self).__init__("DenoiseBregmanFilter")

    def get_filter_padding(self):
        return {}

    def process_frames(self, data):
        data = data[0]
        logging.debug("Running Denoise")
        weight = self.parameters['weight']
        max_iter = self.parameters['max_iterations']
        eps = self.parameters['error_threshold']
        isotropic = self.parameters['isotropic']
        data = np.nan_to_num(data[0, ...])
        result = denoise_tv_bregman(np.float64(data), weight, max_iter=max_iter,
                                    eps=eps, isotropic=isotropic)
        return result

    def get_max_frames(self):
        return 'multiple'



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove objects smaller than the specified size.
   :platform: Unix
   :synopsis: Wrapper around skimage morphology to remove smaller objects

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np
# importing skimage functions
from skimage.morphology import *

@register_plugin
class MorphRemoveObjects(Plugin, CpuPlugin):
    """
    A Plugin to remove objects smaller than the specified size.

    :param min_size: The smallest allowable object size. Default: 64.
    :param connectivity: The connectivity defining the neighborhood of a pixel. Default: 1.
    :param pattern: pattern to apply this to. Default: "VOLUME_XZ".
    """

    def __init__(self):
        super(MorphRemoveObjects, self).__init__("MorphRemoveObjects")

    def setup(self):

        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        out_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

    def pre_process(self):
        # extract given parameters
        self.min_size = self.parameters['min_size']
        self.connectivity = self.parameters['connectivity']

    def process_frames(self, data):
        # run morphological operations here:
        #integerMax = np.max(data[0])
        if (np.sum(data[0]) > 0):
            morph_result = remove_small_objects(data[0].astype(bool), self.min_size, self.connectivity, in_place=False)
            #morph_result = morph_result*1
        else:
            morph_result = np.uint8(np.zeros(np.shape(data[0])))
        return morph_result

    def nInput_datasets(self):
        return 1
    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: yaml_loader
   :platform: Unix
   :synopsis: A class to load data from a folder of Fabio compatible images \
   using data descriptions loaded from a yaml file.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import h5py

from savu.data.data_structures.data_types.image_data import ImageData
from savu.plugins.utils import register_plugin
from savu.plugins.loaders.yaml_converter import YamlConverter


@register_plugin
class ImageTemplateLoader(YamlConverter):
    """
    A class to load data from a folder of FabIO compatible images using data
    descriptions loaded from a yaml file.

    :u*param yaml_file: Path to the file containing the data \
        descriptions. Default: None.
    """

    def __init__(self, name='ImageTemplateLoader'):
        super(ImageTemplateLoader, self).__init__(name)

    def set_data(self, dObj, data):
        folder = data['folder'] if 'folder' in list(data.keys()) else None
        shape = data['shape'] if 'shape' in list(data.keys()) else None

        if not folder:
            raise Exception('Please specify the path to the folder of images.')
        if not shape:
            raise Exception('Please specify the final shape of the data.')

        folder = self.update_value(dObj, folder)
        file_path = self.exp.meta_data.get("data_file")
        dObj.backing_file = h5py.File(file_path, 'r')
        shape = tuple(self.update_value(dObj, shape))
        dObj.data = ImageData(folder, dObj, list(range(len(shape))), shape=shape)
        dObj.set_shape(dObj.data.get_shape())
        return dObj



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove stripe artefacts
   :platform: Unix
   :synopsis: A plugin working in sinogram space to remove stripe artefacts
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
import numpy as np
from scipy.ndimage import gaussian_filter


@register_plugin
class RingRemovalNormalization(Plugin, CpuPlugin):
    """

    Method to remove stripe artefacts in a sinogram (<-> ring artefacts in a \
    reconstructed image) using a normalization-based method. A simple \
    improvement to handle partial stripes is included.

    :param radius: Radius of the Gaussian kernel. Default: 11.
    :param number_of_chunks: Divide the sinogram to many chunks of \
        rows. Default: 1
    """

    def __init__(self):
        super(RingRemovalNormalization, self).__init__(
                "RingRemovalNormalization")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = \
            in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = \
            in_pData[0].get_data_dimension_by_axis_label('rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.width1 = sino_shape[width_dim]
        self.height1 = sino_shape[height_dim]

    def process_frames(self, data):
        sinogram = np.copy(data[0])
        radius = np.clip(np.int16(self.parameters['radius']), 0, self.width1)
        num_chunks = np.clip(np.int16(
                self.parameters['number_of_chunks']), 1, self.height1)
        list_pos = np.array_split(np.arange(self.height1), num_chunks)
        for pos in list_pos:
            bindex = pos[0]
            eindex = pos[-1] + 1
            list_mean = np.mean(sinogram[bindex:eindex, :], axis=0)
            list_mean_filtered = gaussian_filter(list_mean, radius)
            list_coe = list_mean_filtered - list_mean
            mat_coe = \
                np.zeros((eindex - bindex, self.width1), dtype=np.float32)
            mat_coe[:] = list_coe
            sinogram[bindex:eindex, :] = sinogram[bindex:eindex, :] + mat_coe
        return sinogram


# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: test_util_test
   :platform: Unix
   :synopsis: unittest test class for test_utils

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import os
import unittest
import numpy as np
import savu.test.test_utils as tu
from savu.plugins.basic_operations.no_process_plugin import NoProcessPlugin


class MaxFramesTest(unittest.TestCase):

    def __get_slice_list_dict(self, data, pData, pattern, nFrames, dtype,
                              processes):
        data.exp.meta_data.set('processes', processes)
        pData.plugin_data_setup(pattern, nFrames)
        pData._set_meta_data()
        pData._plugin = NoProcessPlugin() # dummy plugin to set required params
        pData.plugin_data_transfer_setup()
        sl_dict = \
            data._get_transport_data()._get_slice_lists_per_process(dtype)
        return sl_dict

    def __assert(self, pData, sl_dict, mft, mfp, nTrans, nProc, nframes=None):
        self.assertEqual(pData._get_max_frames_transfer(), mft)
        self.assertEqual(pData._get_max_frames_process(), mfp)
        self.assertEqual(len(sl_dict['transfer']), nTrans)
        self.assertEqual(len(sl_dict['process']), nProc)

        # checking the distribution of frames is good
        if nframes:
            temp = ((nframes/float(mft)) % 1)
            self.assertTrue((temp > 0.85 or temp == 0.0))

    def __get_system_parameters_file(self):
        path = os.path.join(os.path.dirname(__file__), "system_parameters.yml")
        return path

    def test1_single_threaded(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        params = {'size': (140, 1, 1)}  # data size is (136, 1, 1)
        sys_file = self.__get_system_parameters_file()
        # create a process list?

        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', 'p')
        self.__assert(pData, sl_dict, 17, 1, 8, 17, nframes=136)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', 'p')
        self.__assert(pData, sl_dict, 17, 17, 8, 1, nframes=136)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 8, 'in', 'p')
        self.__assert(pData, sl_dict, 16, 8, 9, 2)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 3, 'in', 'p')
        self.__assert(pData, sl_dict, 18, 3, 8, 6)

    def test2_single_threaded(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        params = {'size': (313, 1, 1)}  # data size is (309, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', 'p')
        self.__assert(pData, sl_dict, 31, 1, 10, 31, nframes=309)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', 'p')
        self.__assert(pData, sl_dict, 31, 31, 10, 1, nframes=309)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 5, 'in', 'p')
        self.__assert(pData, sl_dict, 20, 5, 16, 4)

    def test3_single_threaded(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        params = {'size': (45, 1, 1)}  # data size is (41, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', 'p')
        self.__assert(pData, sl_dict, 21, 1, 2, 21, nframes=41)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', 'p')
        self.__assert(pData, sl_dict, 21, 21, 2, 1, nframes=41)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 17, 'in', 'p')
        self.__assert(pData, sl_dict, 17, 17, 3, 1)

    def test4_single_threaded(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        params = {'size': (4500, 1, 1)}  # data size is (4496, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', 'p')
        self.__assert(pData, sl_dict, 16, 1, 281, 16, nframes=4496)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', 'p')
        self.__assert(pData, sl_dict, 16, 16, 281, 1, nframes=4496)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 6, 'in', 'p')
        self.__assert(pData, sl_dict, 18, 6, 250, 3)

    def test1_parallel(self):
        loader = "random_hdf5_loader"
        params = {}
        processes = ['p']*2        
        params['patterns'] = ['PROJECTION.0s.1s.2c.3c.4s']
        params['axis_labels'] = ['val%d.unit' % i for i in range(5)]
        params['size'] = (4, 3, 1, 1, 4)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))
        data.dtype = np.dtype(np.float32)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 2, 1, 12, 2, nframes=4*3*4)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 2, 2, 12, 1, 4*3*4)

    def test2_parallel(self):
        loader = "random_hdf5_loader"
        params = {}
        processes = ['p']*3        
        params['patterns'] = ['PROJECTION.0s.1s.2c.3c.4s']
        params['axis_labels'] = ['val%d.unit' % i for i in range(5)]
        params['size'] = (1, 1, 1, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))
        data.dtype = np.dtype(np.float32)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 1, 1, 1, 1, 1)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 1, 1, 1, 1, 1)

    def test3_parallel(self):
        loader = "random_hdf5_loader"
        params = {}
        processes = ['p']*4        
        params['patterns'] = ['PROJECTION.0s.1s.2c.3c.4s']
        params['axis_labels'] = ['val%d.unit' % i for i in range(5)]
        params['size'] = (15, 13, 1, 1, 4)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))
        data.dtype = np.dtype(np.float32)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 4, 1, 52, 4, 15*13*4)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 4, 4, 52, 1, 15*13*4)

    def test4_parallel(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        processes = ['p']*20
        params = {'size': (24, 1, 1)}  # data size is (20, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 1, 1, 1, 1, 20)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 1, 1, 1, 1, 20)

    def test5_parallel(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        processes = ['p']*15
        params = {'size': (24, 1, 1)}  # data size is (20, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 2, 1, 1, 2, 20)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 2, 2, 1, 1, 20)

    def test6_parallel(self):
        loader = "full_field_loaders.random_3d_tomo_loader"
        processes = ['p']*15
        params = {'size': (4500, 1, 1)}  # data size is (4496, 1, 1)
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 30, 1, 10, 30, 4496)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 30, 30, 10, 1, 4496)

    def test7_parallel(self):
        loader = "random_hdf5_loader"
        params = {}
        params['patterns'] = ['PROJECTION.0s.1s.2c.3c.4s']
        params['axis_labels'] = ['val%d.unit' % i for i in range(5)]
        params['size'] = (4, 3, 1, 1, 4)
        processes = ['p']*3
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))
        data.dtype = np.dtype(np.float32)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 2, 1, 8, 2, 4*3*4)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 2, 2, 8, 1, 4*3*4)

    def test8_parallel(self):
        loader = "random_hdf5_loader"
        params = {}
        params['patterns'] = ['PROJECTION.0s.1s.2c.3c.4s']
        params['axis_labels'] = ['val%d.unit' % i for i in range(5)]
        params['size'] = (450, 36, 1, 1, 14)
        processes = ['p']*36
        sys_file = self.__get_system_parameters_file()
        data, pData = tu.get_data_object(tu.load_random_data(
                loader, params, system_params=sys_file, fake=True))
        data.dtype = np.dtype(np.float32)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'single', 'in', processes)
        self.__assert(pData, sl_dict, 13, 1, 490, 13)

        sl_dict = self.__get_slice_list_dict(
                data, pData, 'PROJECTION', 'multiple', 'in', processes)
        self.__assert(pData, sl_dict, 13, 13, 490, 1)
#
if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Improve contrast (similar to the Paganin filter)
   :platform: Unix
   :synopsis: A plugin to improve the contrast of the reconstruction image.
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
import numpy as np
import pyfftw.interfaces.scipy_fftpack as fft
from savu.data.plugin_list import CitationInformation


@register_plugin
class FresnelFilter(Plugin, CpuPlugin):
    """
    Method similar to the Paganin filter working both on sinograms and \
    projections. Used to improve the contrast of the reconstruction image.
    :u*param ratio: Control the strength of the filter. Greater is stronger\
    . Default: 100.0
    :u*param pattern: Data processing pattern is 'PROJECTION' or \
        'SINOGRAM'. Default: 'SINOGRAM'.
    """

    def __init__(self):
        super(FresnelFilter, self).__init__("FresnelFilter")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        self.pattern = self.parameters['pattern']
        if self.pattern == "PROJECTION":
            in_pData[0].plugin_data_setup(self.pattern, 'single')
            out_pData[0].plugin_data_setup(self.pattern, 'single')
        else:
            in_pData[0].plugin_data_setup('SINOGRAM', 'single')
            out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def make_window(self, height, width, ratio, pattern):
        center_hei = int(np.ceil((height - 1) * 0.5))
        center_wid = int(np.ceil((width - 1) * 0.5))
        if pattern == "PROJECTION":
            ulist = (1.0 * np.arange(0, width) - center_wid) // width
            vlist = (1.0 * np.arange(0, height) - center_hei) // height
            u, v = np.meshgrid(ulist, vlist)
            win2d = 1.0 + ratio * (u**2 + v**2)
        else:
            ulist = (1.0 * np.arange(0, width) - center_wid) // width
            win1d = 1.0 + ratio * ulist**2
            win2d = np.tile(win1d, (height, 1))
        return win2d

    def apply_filter(self, mat, window, pattern, pad_width):
        (nrow, ncol) = mat.shape
        if pattern == "PROJECTION":
            top_drop = 10  # To remove the time stamp at some data
            mat_pad = np.pad(mat[top_drop:], (
                (pad_width + top_drop, pad_width),
                (pad_width, pad_width)), mode="edge")
            win_pad = np.pad(window, pad_width,
                             mode="edge")
            mat_dec = fft.ifft2(
                fft.fft2(-np.log(mat_pad)) / fft.ifftshift(win_pad))
            mat_dec = np.abs(
                mat_dec[pad_width:pad_width + nrow, pad_width:pad_width + ncol])
        else:
            mat_pad = np.pad(
                -np.log(mat), ((0, 0), (pad_width, pad_width)), mode='edge')
            win_pad = np.pad(window, ((0, 0), (pad_width, pad_width)),
                             mode="edge")
            mat_fft = np.fft.fftshift(fft.fft(mat_pad), axes=1) / win_pad
            mat_dec = fft.ifft(np.fft.ifftshift(mat_fft, axes=1))
            mat_dec = np.abs(mat_dec[:, pad_width:pad_width + ncol])
        return np.float32(np.exp(-mat_dec))

    def pre_process(self):
        inData = self.get_in_datasets()[0]
        self.data_size = inData.get_shape()
        if len(self.data_size) == 4:
            (depth1, height1, width1, _) = self.data_size
        else:
            (depth1, height1, width1) = self.data_size
        ratio = self.parameters['ratio']
        if self.pattern == "PROJECTION":
            self.window = self.make_window(
                height1, width1, ratio, self.pattern)
        else:
            self.window = self.make_window(depth1, width1, ratio, self.pattern)
        self.pad_width = 150

    def process_frames(self, data):
        mat_filt = self.apply_filter(
            data[0], self.window, self.pattern, self.pad_width)
        return mat_filt

    # def get_citation_information(self):
    #     cite_info = CitationInformation()
    #     cite_info.description = \
    #         ("The filter built is based on the Fresnel propagator")
    #     cite_info.bibtex = ()
    #     cite_info.doi = " "
    #     return cite_info
